/*
 * C6Run_mem.c
 */

/*
 * Copyright (C) 2010 Texas Instruments Incorporated - http://www.ti.com/ 
 */

/* 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/************************************************************************
*  Memory allocation routines for C6Run buffers for remote preocedures  *
************************************************************************/


/*  ----------------------------------- Standard Headers              */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdbool.h>

/*  ----------------------------------- Linux Posix Headers           */
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <unistd.h>
#include <pthread.h>

/*  ----------------------------------- CMEM API                      */
#include <ti/sdo/linuxutils/cmem/include/cmem.h>

/*  ----------------------------------- Application headers           */
#include "debug_common.h"
#include "ipc_common.h"
#include "C6Run.h"
#include "dlmalloc.h"

/*  ----------------------------------- This module's header          */
#include "C6Run_mem.h"


/************************************************************
* Explicit External Declarations                            *
************************************************************/


/************************************************************
* Local Macro Declarations                                  *
************************************************************/

#define NUM_TABLE_BINS_BITS           (12)
#define NUM_TABLE_BINS                (0x1 << NUM_TABLE_BINS_BITS)

#define MIN_ALIGNMENT_BITS            (3)
#define MIN_ALIGNMENT                 (0x1 << MIN_ALIGNMENT_BITS)


//#define FAST_BUFFER_LOOKUP


#define C6RUN_MEM_MALLOC_MAGIC        (0x43365231)
#define C6RUN_MEM_MEMALIGN_MAGIC      (0x43365232)

#define SET_MALLOC_MAGIC(P)           (C6RUN_MEM_MALLOC_MAGIC ^ ((uint32_t)P))
#define IS_MALLOC_MAGIC(M,P)          ((((uint32_t)M) ^ ((uint32_t)P)) == C6RUN_MEM_MALLOC_MAGIC)
#define SET_MEMALIGN_MAGIC(P)         (C6RUN_MEM_MEMALIGN_MAGIC ^ ((uint32_t)P))
#define IS_MEMALIGN_MAGIC(M,P)        ((((uint32_t)M) ^ ((uint32_t)P)) == C6RUN_MEM_MEMALIGN_MAGIC)

#define GET_BIN_NUMBER(P)             (((((uint32_t) P) - ((uint32_t)C6RUN_MEM_virtCmemBaseAddr)) >> C6RUN_MEM_binBitShift) & (NUM_TABLE_BINS - 1))

#define C6RUN_MEM_ALIGN(S)            ((((uint32_t)S) + (MIN_ALIGNMENT - 1)) & (~(MIN_ALIGNMENT - 1)))

#define BUF2PTR(B)                    ((void *)((C6RUN_MEM_Buffer *)B + 1))
#define PTR2BUF(P)                    ((pC6RUN_MEM_Buffer)((C6RUN_MEM_Buffer *)P - 1))

#define GET_BUF_SIZE(B)               ((size_t) (((size_t *)B)[-1] & 0xFFFFFFFC) - (2 * sizeof(size_t)))


/************************************************************
* Local Typedef Declarations                                *
************************************************************/


/************************************************************
* Local Function Declarations                               *
************************************************************/

static void LOCAL_insertInBin(pC6RUN_MEM_Buffer buf, void *ptr, size_t size);
static void LOCAL_removeFromBin(pC6RUN_MEM_Buffer buf, void *ptr, size_t size);
static pC6RUN_MEM_Buffer LOCAL_getBuf(void * ptr);


/************************************************************
* Local Variable Definitions                                *
************************************************************/

static CMEM_AllocParams cmem_params = {CMEM_HEAP, CMEM_CACHED, MIN_ALIGNMENT};
static pthread_mutex_t buffTableMutex;


/************************************************************
* Global Variable Definitions                               *
************************************************************/

volatile bool  C6RUN_MEM_initialized    = false;

void          *C6RUN_MEM_virtBaseAddr       = NULL;
void          *C6RUN_MEM_physBaseAddr       = NULL;

void          *C6RUN_MEM_virtCmemBaseAddr   = NULL;
void          *C6RUN_MEM_virtCmemEndAddr    = NULL;
void          *C6RUN_MEM_virtCodeBaseAddr   = NULL;
void          *C6RUN_MEM_virtCodeEndAddr    = NULL;

int32_t        C6RUN_MEM_binBitShift    = 0;
uint32_t       C6RUN_MEM_binSize        = 0;

void          *C6RUN_MEM_bufferTable[NUM_TABLE_BINS];


/************************************************************
* Global Function Definitions                               *
************************************************************/

// Memory subsystem initialization routines
int C6RUN_MEM_init  ( void )
{
  int i;

  if (C6RUN_MEM_initialized)
  {
    return C6RUN_SUCCESS;
  }
  
  VERBOSE_PRINT0("Entered C6RUN_MEM_init()\n");

  // Initialize CMEM
  if (CMEM_init() == -1)
  {  
    printf("CMEM_init() failed!\n");
    return C6RUN_FAIL;
  }
  VERBOSE_PRINT0("CMEM_init() passed!\n");

  // Allocate the entire CMEM region and DSP region.
  // We can perform user-space management in dlmalloc for the CMEM region
  // We can use items from the DSP code region (DSP heap and stack) by doing
  //   an address translation.
  C6RUN_MEM_virtBaseAddr = CMEM_alloc(DSP_REGION_CMEM_SIZE+DSP_REGION_CODE_SIZE, &cmem_params);
  if (C6RUN_MEM_virtBaseAddr == NULL)
  {
    printf("CMEM_alloc() failed!\n");
    return C6RUN_FAIL;
  }
  
  // Figure out how many bits we shift for calculating the correct bin
  // We are rounding up to the nearest power of two and then subdividing that 
  // range of memory into bins.
  for (i = 1; i<(DSP_REGION_CMEM_SIZE); i <<= 1)
  {
    C6RUN_MEM_binBitShift++;
  }
  C6RUN_MEM_binBitShift -= NUM_TABLE_BINS_BITS;
  C6RUN_MEM_binSize = 1 << C6RUN_MEM_binBitShift;
  
  VERBOSE_PRINT1("C6RUN_MEM_binBitShift = %d\n",C6RUN_MEM_binBitShift);
  VERBOSE_PRINT1("C6RUN_MEM_binSize = %d\n",C6RUN_MEM_binSize);  
  
  // Align the CmemVirtBaseAddr to multiple of C6RUN_MEM_binSize
  C6RUN_MEM_virtCmemBaseAddr = (void *) (((intptr_t)C6RUN_MEM_virtBaseAddr + (C6RUN_MEM_binSize - 1)) & (~(C6RUN_MEM_binSize - 1)));
  C6RUN_MEM_virtCmemEndAddr  = (void *) ((intptr_t)C6RUN_MEM_virtCmemBaseAddr) 
                                + DSP_REGION_CMEM_SIZE
                                - ((intptr_t)C6RUN_MEM_virtCmemBaseAddr - (intptr_t)C6RUN_MEM_virtBaseAddr)
                                - 1;
  
  C6RUN_MEM_virtCodeBaseAddr = C6RUN_MEM_virtCmemEndAddr + 1;
  C6RUN_MEM_virtCodeEndAddr  = C6RUN_MEM_virtCodeBaseAddr + DSP_REGION_CODE_SIZE - 1;
  
  C6RUN_MEM_physBaseAddr  = (void *)CMEM_getPhys(C6RUN_MEM_virtBaseAddr);
  VERBOSE_PRINT0("CMEM_alloc() passed!\n");
  
  // Init the C6RUN_MEM_bufferTable to NULL values
  memset(C6RUN_MEM_bufferTable,0x00,(NUM_TABLE_BINS*sizeof(void *)));

  // Init the mutex protecting the bufferTable
  pthread_mutex_init(&buffTableMutex,NULL);

  C6RUN_MEM_initialized = true;

  // Register memory exit routine for calling upon program termination
  atexit(C6RUN_MEM_exit);
  
  VERBOSE_PRINT0("Leaving C6RUN_MEM_init()\n");
  
  return C6RUN_SUCCESS;
}

void C6RUN_MEM_exit ( void )
{
  if (!C6RUN_MEM_initialized)
  {
    return;
  }
  
  VERBOSE_PRINT0("Entered C6RUN_MEM_exit()\n");

  // Check for orphaned memory(leaks) and print it out
  for (int i=0; i<NUM_TABLE_BINS; i++)
  {
    void *ptr = C6RUN_MEM_bufferTable[i];
    while (ptr != NULL)
    {
      VERBOSE_PRINT2("Orphaned memory at bin %d, 0x%08X.\n",i,(intptr_t)ptr);
      ptr = (LOCAL_getBuf(ptr))->next;
    }
  }
  
  // Free up the entire CMEM region
  CMEM_free(C6RUN_MEM_virtBaseAddr,&cmem_params);
  
  if (CMEM_exit() < 0)
  {
    printf("CMEM_exit() failed!\n");
  }
  VERBOSE_PRINT0("CMEM_exit() passed!\n");

  C6RUN_MEM_initialized = false;
  
  VERBOSE_PRINT0("Leaving C6RUN_MEM_exit()\n");
}


// C6Run memory buffer management routines
void *C6RUN_MEM_malloc(size_t size)
{
  void *tmpPtr;
  
  VERBOSE_PRINT1("C6RUN_MEM_malloc(%d)\n",size);
  
  tmpPtr = C6RUN_MEM_memalign(MIN_ALIGNMENT,size);

  VERBOSE_PRINT1("C6RUN_MEM_malloc returns = 0x%08X\n",(uint32_t)tmpPtr);
  
  return tmpPtr;
}

void *C6RUN_MEM_realloc(void *oldPtr,size_t size)
{
  size_t totalSize;
  void *tmpPtr = NULL;
  pC6RUN_MEM_Buffer pBuffer;
  
  VERBOSE_PRINT2("C6RUN_MEM_realloc(0x%08X,%d)\n",(uint32_t)oldPtr,size);
  
  if (!C6RUN_MEM_initialized)
  {
    C6RUN_MEM_init();
  }
  
  if (oldPtr == NULL)
  {
    return C6RUN_MEM_malloc(size);
  }
  
  // Calculate total size
  totalSize = (size + sizeof(C6RUN_MEM_Buffer));
  totalSize = C6RUN_MEM_ALIGN(totalSize);
  
  // Get the buffer
  pBuffer = LOCAL_getBuf(oldPtr);
  if (pBuffer == NULL)
  {
    return NULL;
  }
  
  // Realloc the buffer
  tmpPtr = dlrealloc((void *)pBuffer,totalSize);
  
  if (tmpPtr)
  {
    pBuffer = (pC6RUN_MEM_Buffer) tmpPtr;
    tmpPtr = BUF2PTR(pBuffer);
    // If the new pointer is not the same as the old one, we have
    // some updates we need to do in the buffer table
    if (tmpPtr != oldPtr)
    {
      C6RUN_MEM_free(oldPtr);
      
      pBuffer->next = NULL;
      pBuffer->magic = SET_MALLOC_MAGIC(tmpPtr);
      
      pthread_mutex_lock(&buffTableMutex);       
      
      // Add new pointer/buffer to table
      LOCAL_insertInBin(pBuffer, tmpPtr, size);
      
      pthread_mutex_unlock(&buffTableMutex);
    }
  }
  
  VERBOSE_PRINT1("C6RUN_MEM_realloc returns = 0x%08X\n",(uint32_t)tmpPtr);
  
  return tmpPtr;
}

void *C6RUN_MEM_memalign(size_t alignment, size_t size)
{
  size_t totalSize;
  void *buf = NULL, *ptr = NULL;
  uint32_t startMemPtr, endMemPtr;
  
  VERBOSE_PRINT2("C6RUN_MEM_memalign(%d,%d)\n",alignment,size);

  if (!C6RUN_MEM_initialized)
  {
    C6RUN_MEM_init();
  }
  
  // Check for non power-of-two alignment
  if ((alignment & (alignment - 1)))
  {
    VERBOSE_PRINT0("C6RUN_MEM_memalign: Error, invalid alignment!\n");
    return NULL;
  }
  
  // Aligment is at least MIN_ALIGNMENT<<1 bytes
  totalSize   = (alignment-MIN_ALIGNMENT) + size + sizeof(C6RUN_MEM_Buffer);
  totalSize   = C6RUN_MEM_ALIGN(totalSize);
  
  // Get the allocated data from the dlmalloc implementation
  buf = dlmalloc(totalSize);
  if (buf == NULL)
  {
    return NULL;
  }
    
  startMemPtr = (uint32_t) BUF2PTR(buf);
  endMemPtr   = (uint32_t) (startMemPtr + size - 1);  
  
  if (GET_BIN_NUMBER(startMemPtr) != GET_BIN_NUMBER(endMemPtr))
  {
    // If the buffer spans multiple bins then we need to align
    // our buffer so that the startMemPtr sits at the start of the 
    // next bin
    VERBOSE_PRINT1("Bin spanning buffer at 0x%08X\n",(uint32_t)buf);
  
    // Now resize the buffer previously malloced
    dlfree(buf);
    buf = dlmemalign(C6RUN_MEM_binSize, totalSize);
    if (buf == NULL)
    {
      return NULL;
    }
    VERBOSE_PRINT1("  realigning to start of next bin at 0x%08X\n",(uint32_t)buf);
  }

  if (buf)
  {
    pC6RUN_MEM_Buffer pBuffer = (pC6RUN_MEM_Buffer) buf;
    intptr_t bufPtr;

    pBuffer->next = NULL;
    ptr = BUF2PTR(buf);
    pBuffer->magic = SET_MALLOC_MAGIC(ptr);

    bufPtr = (intptr_t)ptr;
    if (bufPtr & (alignment - 1))
    {
      // Align the pointer
      bufPtr += ((-bufPtr) & (alignment - 1));
      
      ptr = (void *)bufPtr;
    }
  
    if (BUF2PTR(buf) != ptr)
    {
      // There is always enough space for the base pointer and the memalign magic word
      ((uint32_t*)ptr)[-1] = SET_MEMALIGN_MAGIC(ptr);
      ((void**)ptr)[-2] = buf;
    }
  }
  
  if (ptr)
  {
    // Insert the pBuffer into the buffer table
    pthread_mutex_lock(&buffTableMutex);    
    
    LOCAL_insertInBin(buf, ptr, size);
    
    pthread_mutex_unlock(&buffTableMutex);
  }
  
  VERBOSE_PRINT1("C6RUN_MEM_memalign returns = 0x%08X\n",(uint32_t)ptr);

  return ptr;
}

void *C6RUN_MEM_calloc(size_t nelem, size_t elsize)
{
  size_t totalSize;
  void*  tmpPtr;
  
  VERBOSE_PRINT2("C6RUN_MEM_calloc(%d,%d)\n",nelem,elsize);
  
  totalSize = nelem * elsize;
  tmpPtr  = C6RUN_MEM_memalign(MIN_ALIGNMENT,totalSize);
  if (tmpPtr != NULL)
  {
    memset(tmpPtr, 0x00, totalSize);
  }
  
  VERBOSE_PRINT1("C6RUN_MEM_calloc returns = 0x%08X\n",(uint32_t)tmpPtr);
  
  return tmpPtr;
}

void C6RUN_MEM_free(void* ptr)
{
  
  VERBOSE_PRINT1("C6RUN_MEM_free(0x%08X)\n",(uint32_t)ptr);

  if (!C6RUN_MEM_initialized)
  {
    C6RUN_MEM_init();
  }
  
  if (ptr)
  {
    pC6RUN_MEM_Buffer pBuffer = LOCAL_getBuf(ptr);
    
    if (pBuffer)
    {
      size_t size = GET_BUF_SIZE(pBuffer) - (((intptr_t)ptr) - ((intptr_t)pBuffer));

      pthread_mutex_lock(&buffTableMutex);
    
      LOCAL_removeFromBin(pBuffer, ptr, size);
    #if defined(FAST_BUFFER_LOOKUP)
      pBuffer->next   = NULL;
      pBuffer->magic  = NULL;
    #endif
      dlfree((void *)pBuffer);
    
      pthread_mutex_unlock(&buffTableMutex); 
    }
  }
  
  VERBOSE_PRINT0("Leaving C6RUN_MEM_free.\n");
}


// Memory cache management routines
void C6RUN_MEM_globalInv()
{
  CMEM_cacheInv(C6RUN_MEM_virtBaseAddr, DSP_REGION_CMEM_SIZE+DSP_REGION_CODE_SIZE);
}

void C6RUN_MEM_globalWb()
{
  CMEM_cacheWb(C6RUN_MEM_virtBaseAddr, DSP_REGION_CMEM_SIZE+DSP_REGION_CODE_SIZE);
}

void C6RUN_MEM_inv(void *ptr,size_t size)
{
  CMEM_cacheInv(ptr, size);
}

void C6RUN_MEM_wb(void *ptr,size_t size)
{
  CMEM_cacheWb(ptr, size);
}

void C6RUN_MEM_wbInv(void *ptr,size_t size)
{
  CMEM_cacheWbInv(ptr, size);
}

void C6RUN_MEM_globalWbInv()
{
  CMEM_cacheWbInv(C6RUN_MEM_virtBaseAddr, DSP_REGION_CMEM_SIZE+DSP_REGION_CODE_SIZE);
}


// Memory helper routines
void C6RUN_MEM_printStats(void)
{
  uint32_t usedBinCount = 0, lastUsedBin=0;
  uint32_t totalUsedMemory = 0, binUsedMemory = 0;
  int i;
  
  // Make sure the memory system is initialized
  if (!C6RUN_MEM_initialized)
  {
    C6RUN_MEM_init();
  }
  
  printf("C6RUN_MEM Statistics:\n");
  
  for (i = 0; i<NUM_TABLE_BINS; ++i)
  {
    void *ptr = C6RUN_MEM_bufferTable[i];
    if (ptr)
    {
      ++usedBinCount;
      lastUsedBin = i;
      binUsedMemory = 0;
      
      while (ptr)
      {
        pC6RUN_MEM_Buffer buf = LOCAL_getBuf(ptr);
        binUsedMemory += GET_BUF_SIZE(buf) - (((intptr_t)ptr) - ((intptr_t)buf));
        ptr = buf->next;
      }
      totalUsedMemory += binUsedMemory;
      printf("\tBin %d: Memory used = 0x%08X\n",i,binUsedMemory);
    }
  }
 
  printf("\n\tLast Used Bin = %d\n",lastUsedBin);
  printf("\tTotal Bins Used = %d\n",usedBinCount);
  printf("\tTotal memory used = 0x%08X\n",totalUsedMemory);
}

void *C6RUN_MEM_lookupBuffer(void *ptr, size_t *buffSize)
{
  uint32_t binNum;
  void *currPtr;
  uint32_t base = 0, size = 0;
  bool bufferFound = false;
  pC6RUN_MEM_Buffer pBuf = NULL;
  
  VERBOSE_PRINT1("C6RUN_MEM_lookupBuffer(0x%08X,...)\n",(uint32_t)ptr);

  binNum = GET_BIN_NUMBER(ptr);
  currPtr = C6RUN_MEM_bufferTable[binNum];

  VERBOSE_PRINT1("C6RUN_MEM_lookupBuffer: binNum = %d \n",binNum);
  
  if (ptr)
  {
  #if defined(FAST_BUFFER_LOOKUP)
    pBuf = LOCAL_getBuf(ptr);
    if (pBuf != NULL)
    {
      VERBOSE_PRINT0("C6RUN_MEM_lookupBuffer: pBuf fast lookup succeeded.\n");
    
      *buffSize = GET_BUF_SIZE(pBuf) - (((uint32_t)ptr) - ((uint32_t)pBuf));
      return ((void *)ptr);  
    }
  #endif  
  
    // Check to make sure we have a populated bin
    if (currPtr)
    {
      // Search through buffer table bin for a match. A match is the
      // input pointer inside the range of an allocated buffer
      while ((currPtr != NULL) && (!bufferFound))
      {
        base = (uint32_t) currPtr;
        pBuf = LOCAL_getBuf(currPtr);
        
        if (pBuf == NULL)
        {
          VERBOSE_PRINT0("C6RUN_MEM_lookupBuffer: pBuf lookup failed.\n");
          exit(0);
        }
      
        // Get size from dlmalloc header data (chunk size minus dlmalloc 
        // header size minus C6RUN_MEM_Buffer size minus any filler)
        size = GET_BUF_SIZE(pBuf) - (((uint32_t)currPtr) - ((uint32_t)pBuf));
      
        if ((base <= ((uint32_t) ptr)) && (((uint32_t) ptr) < (base + size)))
        {
          VERBOSE_PRINT1("  Found Buffer : 0x%08X\n",(uint32_t)pBuf);
          bufferFound = true;
          break;
        }
        
        VERBOSE_PRINT1("  Skip Buffer : 0x%08X\n",(uint32_t)pBuf);
        
        currPtr = pBuf->next;
      }
      
      if (!bufferFound)
      {
        size = 0;
        base = 0;
      }
    }
  }
  
  VERBOSE_PRINT1("C6RUN_MEM_lookupBuffer returns base   = 0x%08X \n",base);
  VERBOSE_PRINT1("C6RUN_MEM_lookupBuffer returns size   = 0x%08X \n",size);
  
  *buffSize = size;
  return ((void *)base);
}

void *C6RUN_MEM_virt2Phys(void *virtPtr)
{
  return (void *) (((virtPtr) - C6RUN_MEM_virtBaseAddr) + C6RUN_MEM_physBaseAddr);
}

void *C6RUN_MEM_phys2Virt(void *physPtr)
{
  return (void *) (((physPtr) - C6RUN_MEM_physBaseAddr) + C6RUN_MEM_virtBaseAddr);
}


/***********************************************************
* Local Function Definitions                               *
***********************************************************/

static void LOCAL_insertInBin(pC6RUN_MEM_Buffer buf, void *ptr, size_t size)
{
  uint32_t binNum, binCount, totalBinCount;
  
  if (buf == NULL)
  {
    return;
  }
  
  // Get the binNumber
  binNum = GET_BIN_NUMBER(ptr);
  
  VERBOSE_PRINT1("Insert - Bin: %d\n",binNum);
  VERBOSE_PRINT1("  Buffer:  0x%08X\n",(uint32_t)buf);
  VERBOSE_PRINT1("  Pointer: 0x%08X\n",(uint32_t)ptr);

  // Round the size variable up to the next multiple of the binSize
  size = (size + (C6RUN_MEM_binSize - 1)) & ~(C6RUN_MEM_binSize - 1);
  totalBinCount = size >> C6RUN_MEM_binBitShift;
  binCount = 0;
  
  // For every bin but the last, just set the bin to the buffer.
  // We can do this since we know each bin is full with this one buffer.
  while ((++binCount) < totalBinCount)
  {
    C6RUN_MEM_bufferTable[binNum] = ptr;
    binNum++;
  }

  // For the last buffer, we might have other ones already in that bin
  // so we need to handle this appropriatley
  buf->next = C6RUN_MEM_bufferTable[binNum];
  C6RUN_MEM_bufferTable[binNum] = ptr;
}

static void LOCAL_removeFromBin(pC6RUN_MEM_Buffer buf, void *ptr, size_t size)
{
  uint32_t binNum, binCount, totalBinCount;
  pC6RUN_MEM_Buffer pCurrBuffer, pNextBuffer;
  
  if ((ptr == NULL) || (buf == NULL) || (size == 0))
  {
    return;
  }
  
  // Get the binNumber
  binNum = GET_BIN_NUMBER(ptr);
  
  VERBOSE_PRINT2("Remove - Bin: %d,Pointer: 0x%08X\n",binNum,(uint32_t)ptr);
  
  // Round the size variable up to the next multiple of the binSize
  size = (size + (C6RUN_MEM_binSize - 1)) & ~(C6RUN_MEM_binSize - 1);
  totalBinCount = size >> C6RUN_MEM_binBitShift;
  binCount = 1;
  
  // For every bin but the last, just set the bin to the NULL.
  // We can do this since we know each bin is used by this one buffer.
  while (binCount < totalBinCount)
  {
    C6RUN_MEM_bufferTable[binNum] = NULL;
    binNum++;
    binCount++;
  }

  // For the last buffer, we might have other ones still in the bin
  // so we need to handle this case appropriately 
  
  // Simply go through the bin list and remove this one
  pCurrBuffer = LOCAL_getBuf(C6RUN_MEM_bufferTable[binNum]);
  if (ptr == C6RUN_MEM_bufferTable[binNum])
  {
    if (pCurrBuffer)
    {
      C6RUN_MEM_bufferTable[binNum] = pCurrBuffer->next;
      VERBOSE_PRINT1("  Found Buffer : 0x%08X\n",(uint32_t)pCurrBuffer);
    }
    else
    {
      VERBOSE_PRINT0("  Bin buffer mismatch: \n");
      VERBOSE_PRINT1("    Buffer : 0x%08X\n",(uint32_t)pCurrBuffer);
    }
  }
  else
  {
    while (pCurrBuffer->next != NULL)
    {
      pNextBuffer = LOCAL_getBuf(pCurrBuffer->next);
      if (pNextBuffer)
      {
        if (ptr == pCurrBuffer->next)
        {
          pCurrBuffer->next = pNextBuffer->next;
          VERBOSE_PRINT1("  Skip Buffer : 0x%08X\n",(uint32_t)pCurrBuffer);
          VERBOSE_PRINT1("  Found Buffer: 0x%08X\n",(uint32_t)pNextBuffer);
          break;
        }
        VERBOSE_PRINT1("  Skip Buffer : 0x%08X\n",(uint32_t)pCurrBuffer);
        pCurrBuffer = pNextBuffer;
      }
      else
      {
        VERBOSE_PRINT0("  Bin buffer mismatch: \n");
        VERBOSE_PRINT1("    Buffer : 0x%08X\n",(uint32_t)pNextBuffer);
        break;
      }
    }
  }
}

static pC6RUN_MEM_Buffer LOCAL_getBuf(void * ptr)
{
  pC6RUN_MEM_Buffer pBuffer = PTR2BUF(ptr);
  
  if (!IS_MALLOC_MAGIC(pBuffer->magic,ptr))
  {
    if ( IS_MEMALIGN_MAGIC((((uint32_t *)ptr)[-1]),ptr) )
    {
      // It appears to be a memaligned buffer
      pBuffer = ((void**)ptr)[-2];
      
      ptr = BUF2PTR(pBuffer);
      if (!IS_MALLOC_MAGIC(pBuffer->magic, ptr))
      {
        VERBOSE_PRINT0("LOCAL_getBuf() failed because memalign magic number found, but no malloc magic number found: \n");
        VERBOSE_PRINT1("  pBuffer        = 0x%08X \n",(unsigned int) pBuffer);
        VERBOSE_PRINT1("  ptr            = 0x%08X \n",(unsigned int) ptr);
        VERBOSE_PRINT1("  pBuffer->magic = 0x%08X \n",(unsigned int) pBuffer->magic ^ ((unsigned int)ptr));
        pBuffer = NULL;
      }
    }
    else
    {
      VERBOSE_PRINT0("LOCAL_getBuf() failed because no magic number found: \n");
      VERBOSE_PRINT1("  pBuffer        = 0x%08X \n",(unsigned int) pBuffer);      
      VERBOSE_PRINT1("  ptr            = 0x%08X \n",(unsigned int) ptr);
      VERBOSE_PRINT1("  ptr[-1]        = 0x%08X \n",(((uint32_t *)ptr)[-1]));
      VERBOSE_PRINT1("  ptr[-2]        = 0x%08X \n",(((uint32_t *)ptr)[-2]));
      pBuffer = NULL;
    }
  }
  
  // Do some additional checks to make sure the above checks are not returning
  // a false positive
#if defined(FAST_BUFFER_LOOKUP)
  if (pBuffer)
  {
    if (pBuffer->next)
    {
      // Make sure the pBuffer->next is in the valid range
      if ( ( ((uint32_t)pBuffer->next) < ((uint32_t)C6RUN_MEM_virtCmemBaseAddr) ) ||
           ( ((uint32_t)pBuffer->next) > ((uint32_t)C6RUN_MEM_virtCmemEndAddr) ) )
      {
        return NULL;
      }
    }
    
    if (GET_BUF_SIZE(pBuffer) >= (DSP_REGION_CMEM_SIZE + DSP_REGION_CODE_SIZE))
    {
      return NULL;
    }
  }
#endif
  
  return pBuffer;
}

/***********************************************************
* End file                                                 *
***********************************************************/
