/*
 * control_ipc.c
 */

/*
 * Copyright (C) 2010 Texas Instruments Incorporated - http://www.ti.com/ 
 */

/* 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/************************************************************************
*  Routines for handling C I/O requests from the DSP                    *
************************************************************************/

/*  ----------------------------------- Standard Headers                */
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <time.h>

/*  ----------------------------------- Linux Posix Headers            */
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <unistd.h>

/*  ----------------------------------- SysLink                         */
#include <ti/syslink/Std.h>
#include <ti/syslink/SysLink.h>
#include <ti/syslink/IpcHost.h>

/*  ----------------------------------- SysLink IPC                     */
#include <ti/ipc/Ipc.h>
#include <ti/ipc/MultiProc.h>
#include <ti/ipc/HeapMemMP.h>
#include <ti/ipc/SharedRegion.h>
#include <ti/ipc/MessageQ.h>
#include <ti/ipc/Notify.h>

/*  ----------------------------------- CMEM API                        */
#include <ti/sdo/linuxutils/cmem/include/cmem.h>

/*  ----------------------------------- Application headers             */
#include "debug_common.h"
#include "C6Run.h"
#include "C6Run_mem.h"
#include "cio_ipc.h"
#include "rpc_ipc.h"

/*  ----------------------------------- This module's header            */
#include "control_ipc.h"


/************************************************************
* Explicit External Declarations                            *
************************************************************/


/************************************************************
* Local Macro Declarations                                  *
************************************************************/


/************************************************************
* Local Typedef Declarations                                *
************************************************************/


/************************************************************
* Local Function Declarations                               *
************************************************************/


/************************************************************
* Local Variable Definitions                                *
************************************************************/


/************************************************************
* Global Variable Definitions                               *
************************************************************/

volatile bool CONTROL_IPC_initialized  = FALSE;

// Global CONTROL MessageQ Objects for Tx and Rx
MessageQ_Handle   MessageQ_controlToGPP;
MessageQ_QueueId  MessageQ_controlFromGPPId;

CONTROL_Msg *CONTROL_msg;


/************************************************************
* Global Function Definitions                               *
************************************************************/

int32_t CONTROL_IPC_init( void )
{
  int32_t status = C6RUN_SUCCESS;

  if (CONTROL_IPC_initialized)
  {
    return C6RUN_SUCCESS;
  }
  
  VERBOSE_PRINT0("Entered CONTROL_IPC_init()\n");

  // Create the local receive message queue (CONTROL TO GPP queue)
  MessageQ_controlToGPP = MessageQ_create (CONTROL_TO_GPP_MSGQNAME, NULL);
  if (MessageQ_controlToGPP == NULL)
  {
    status = MessageQ_E_FAIL;
    printf ("MessageQ_create() failed: status = [0x%x] \n", status);
    return C6RUN_FAIL;
  }

  // Connect to the remote sending message queue
  do
  {
    status = MessageQ_open (CONTROL_FROM_GPP_MSGQNAME, &MessageQ_controlFromGPPId);
  }
  while (status == MessageQ_E_NOTFOUND);
  
  if (status < 0)
  {
    printf ("MessageQ_open() failed: status = [0x%x]\n", status);
    return C6RUN_FAIL;
  }
  
  CONTROL_msg = (CONTROL_Msg *) MessageQ_alloc(COMMON_HEAP_ID, sizeof(CONTROL_Msg));
  if (CONTROL_msg == NULL)
  { 
    REPORT_STATUS("MessageQ_alloc");
    return C6RUN_FAIL;
  }

  CONTROL_IPC_initialized = TRUE;
  
  VERBOSE_PRINT0("Leaving CONTROL_IPC_init()\n");
  
  return C6RUN_SUCCESS;
}

void CONTROL_IPC_exit()
{
  int32_t status = 0;
  int32_t tmpStatus = 0;
  
  if (!CONTROL_IPC_initialized)
  {
    return;
  }

  VERBOSE_PRINT0("Entered CONTROL_IPC_exit()\n");

  MessageQ_free((MessageQ_Msg) CONTROL_msg);
  
  status = MessageQ_close(&MessageQ_controlFromGPPId);
  if (status < 0)
  {
    printf ("MessageQ_close() failed: status = [0x%x]\n", status);
  }
  else
  {
    VERBOSE_PRINT0("MessageQ_close() passed.\n");
  }
  
  tmpStatus = MessageQ_delete(&MessageQ_controlToGPP);
  if ((status >= 0) && (tmpStatus < 0))
  {
    status = tmpStatus;
    printf ("MessageQ_delete() failed: status = [0x%x]\n", status);
  }
  else
  {
    VERBOSE_PRINT0("MessageQ_delete() passed.\n");
  }
  
  CONTROL_IPC_initialized = FALSE;    

  VERBOSE_PRINT0("Leaving CONTROL_IPC_exit()\n");
}

void CONTROL_IPC_send(uint16_t command)
{
  int32_t status;
  
  VERBOSE_PRINT0("Entering CONTROL_IPC_send()\n");

  MessageQ_setMsgId(CONTROL_msg, command);

  status = MessageQ_put(MessageQ_controlFromGPPId, (MessageQ_Msg) CONTROL_msg);
  
  if (status < 0)
  { 
    printf ("MessageQ_put() failed: status = [0x%x]\n", status); 
  }
  
  VERBOSE_PRINT0("Exiting CONTROL_IPC_send()\n");
}

uint16_t CONTROL_IPC_recv()
{
  int32_t status = 0;
  uint16_t msgId = 0;
  
  VERBOSE_PRINT0("Entering CONTROL_IPC_recv()\n");

  status = MessageQ_get(MessageQ_controlToGPP, (MessageQ_Msg *) &CONTROL_msg, MessageQ_FOREVER);

  if (status < 0)
  {
    printf ("MessageQ_get() failed: status = [0x%x]\n", status); 
    return CONTROL_NULL;
  }

  msgId = MessageQ_getMsgId(CONTROL_msg);

  VERBOSE_PRINT0("Exiting CONTROL_IPC_recv()\n");
  
  return msgId;
}


/***********************************************************
* Local Function Definitions                               *
***********************************************************/


/***********************************************************
* End file                                                 *
***********************************************************/

