/*
 * rpc_ipc.c
 */

/*
 * Copyright (C) 2010 Texas Instruments Incorporated - http://www.ti.com/ 
 */

/* 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/************************************************************************
*  Routines for handling C I/O requests from the DSP                    *
************************************************************************/

/*  ----------------------------------- Standard Headers                */
#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <time.h>

/*  ----------------------------------- Linux Posix Headers            */
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <unistd.h>
#include <semaphore.h>

/*  ----------------------------------- SysLink                         */
#include <ti/syslink/Std.h>
#include <ti/syslink/SysLink.h>

/*  ----------------------------------- SysLink API                     */
#include <ti/syslink/IpcHost.h>
#include <ti/syslink/ProcMgr.h>
#include <ti/ipc/MultiProc.h>

/*  ----------------------------------- Application headers             */
#include "debug_common.h"
#include "C6Run.h"
#include "C6Run_mem.h"

/*  ----------------------------------- This module's header            */
#include "rpc_ipc.h"


/************************************************************
* Explicit External Declarations                            *
************************************************************/


/************************************************************
* Local Macro Declarations                                  *
************************************************************/


/************************************************************
* Local Typedef Declarations                                *
************************************************************/


/************************************************************
* Local Function Declarations                               *
************************************************************/


/************************************************************
* Local Variable Definitions                                *
************************************************************/

static sem_t RPC_IPC_msgSemObj;

static bool LOCAL_isModuleInitialized = false;


/************************************************************
* Global Variable Definitions                               *
************************************************************/

// Global CONTROL MessageQ Objects for Tx and Rx
MessageQ_Handle   MessageQ_rpcToGPP;
MessageQ_QueueId  MessageQ_rpcFromGPPId;


RPC_Msg *RPC_msg;


/************************************************************
* Global Function Definitions                               *
************************************************************/

int32_t RPC_IPC_init()
{
  int32_t status = 0;

  if (LOCAL_isModuleInitialized)
  {
    return C6RUN_SUCCESS;
  }
  
  VERBOSE_PRINT0("Entered RPC_IPC_init()\n");

  // Create the local receive message queue (RPC TO GPP queue)
  MessageQ_rpcToGPP = MessageQ_create (RPC_TO_GPP_MSGQNAME, NULL);
  if (MessageQ_rpcToGPP == NULL)
  {
    status = MessageQ_E_FAIL;
    printf ("MessageQ_create() failed: status = [0x%8X] \n", status);
    return C6RUN_FAIL;
  }

  // Connect to the remote sending message queue
  do
  {
    status = MessageQ_open (RPC_FROM_GPP_MSGQNAME, &MessageQ_rpcFromGPPId);
  }
  while (status == MessageQ_E_NOTFOUND);
  
  if (status < 0)
  {
    printf ("MessageQ_open() failed: status = [0x%8X]\n", status);
    return C6RUN_FAIL;
  }
  
  if( sem_init(&RPC_IPC_msgSemObj,0,RPC_MSG_CNT) < 0 )
  {
    printf("sem_init() failed");
    exit(0);
  }
  
  LOCAL_isModuleInitialized = true;
  
  VERBOSE_PRINT0("Leaving RPC_IPC_init()\n");
  
  return C6RUN_SUCCESS;
}

void RPC_IPC_exit()
{
  int32_t status      = 0;
  int32_t tmpStatus   = 0;
  
  if (!LOCAL_isModuleInitialized)
  {
    return;
  }

  VERBOSE_PRINT0("Entered RPC_IPC_exit()\n");

  status = MessageQ_close(&MessageQ_rpcFromGPPId);
  if (status < 0)
  {
    printf ("MessageQ_close() failed: status = [0x%8X]\n", status);
  }
  else
  {
    VERBOSE_PRINT0("MessageQ_close() passed.\n");
  }
  
  tmpStatus = MessageQ_delete(&MessageQ_rpcToGPP);
  if ((status >= 0) && (tmpStatus < 0))
  {
    printf ("MessageQ_delete() failed: status = [0x%8X]\n", status);
  }
  else
  {
    VERBOSE_PRINT0("MessageQ_delete() passed.\n");
  }    

  LOCAL_isModuleInitialized = false;

  VERBOSE_PRINT0("Leaving RPC_IPC_exit()\n");
}

void RPC_IPC_sendMsg( RPC_Msg *msg, uint16_t msgId )
{
  int32_t status;
  
  VERBOSE_PRINT0("Entering RPC_IPC_sendMsg()\n");

  MessageQ_setMsgId(msg, msgId);

  status = MessageQ_put(MessageQ_rpcFromGPPId, (MessageQ_Msg) msg);
  
  if (status < 0)
  { 
    printf ("MessageQ_put() failed: status = [0x%8X]\n", status); 
  }  
  
  VERBOSE_PRINT0("Exiting RPC_IPC_sendMsg()\n");
}

uint16_t RPC_IPC_recvMsg( RPC_Msg **msg)
{
  int32_t status;
  uint16_t msgId;
  
  VERBOSE_PRINT0("Entering RPC_IPC_recvMsg()\n");

  status = MessageQ_get(MessageQ_rpcToGPP, (MessageQ_Msg *) msg, MessageQ_FOREVER);

  if (status < 0)
  {
    printf ("MessageQ_get() failed: status = [0x%8X]\n", status); 
    return RPC_NULL;
  }

  msgId = MessageQ_getMsgId((*msg));
  
  VERBOSE_PRINT0("Exiting RPC_IPC_recvMsg()\n");
  
  return msgId;
}

void RPC_IPC_allocMsg( RPC_Msg **msg )
{

  VERBOSE_PRINT0("Entering RPC_IPC_allocMsg()\n");

  // Semaphore wait on an available message
  sem_wait(&RPC_IPC_msgSemObj);

  {
    // Do MessageQ alloc for a remote procedure call
    *msg = (RPC_Msg *) MessageQ_alloc(COMMON_HEAP_ID, sizeof(RPC_Msg));
    if (*msg == NULL)
    { 
      printf ("MessageQ_alloc() failed.");
    }    
  }  
  
  VERBOSE_PRINT0("Exiting RPC_IPC_allocMsg()\n");
}

void RPC_IPC_freeMsg( RPC_Msg *msg )
{

  VERBOSE_PRINT0("Entering RPC_IPC_freeMsg()\n");

  MessageQ_free((MessageQ_Msg) msg);

  sem_post(&RPC_IPC_msgSemObj);
  
  VERBOSE_PRINT0("Exiting RPC_IPC_freeMsg()\n");
}


/***********************************************************
* Local Function Definitions                               *
***********************************************************/


/***********************************************************
* End file                                                 *
***********************************************************/

